
/*
Exemplo cliente-servidor
Codigo do cliente
(C) 2011-2012 Pedro Freire

Compilar (em Mac/Unix) com
	gcc -Wall client.c -o client
ou em Windows com
	gcc -Wall client.c -o client -lws2_32

Correr com
	client <fqdn-do-servidor>:<porta-do-servidor>
ou
	client <fqdn-do-servidor>
ou
	client
(esta ultima liga-se a um servidor a correr no mesmo computador e
que foi arrancado sem argumentos, ou seja, com a porta de omissao)
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#ifdef errno
#undef errno
#endif
#define EADDRNOTAVAIL		WSAEADDRNOTAVAIL
#define EADDRINUSE		WSAEADDRINUSE
#define ETIMEDOUT		WSAETIMEDOUT
#define ECONNREFUSED		WSAECONNREFUSED
#define ECONNRESET		WSAECONNRESET
#define perror_and_exit(msg)	{ fprintf(stderr, "Erro %i do %s do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms740668(v=vs.85).aspx\npara saber a descricao.\n", (int) WSAGetLastError(), msg); exit(EXIT_FAILURE); }
#define errno			WSAGetLastError()
#define sleep(s)		Sleep((s)*1000)
typedef int socklen_t;
#define IPPORT_USERRESERVED	5000
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#define INVALID_SOCKET		-1
#define SOCKET_ERROR		-1
#define perror_and_exit(msg)	{ perror(msg); exit(EXIT_FAILURE); }
#endif


#define DEFAULT_HOST	"localhost"
#define DEFAULT_PORT	IPPORT_USERRESERVED


/* Envia o ficheiro com o nome "filename" para o servidor no socket
   "server_socket".
   Devolve true (1) com sucesso, false (0) se o servidor se desligou.
*/
int send_file( int server_socket, char *filename )
{
	char         buffer[1024+1];
	FILE         *fp;
	struct stat  s;
	int          size, send_sz;

	fp = fopen( filename, "r" );
	if( fp == NULL )
		{
		perror( "fopen()" );
		exit( EXIT_FAILURE );
		}
	if( fstat(fileno(fp), &s) == -1 )
		{
		perror( "fstat()" );
		exit( EXIT_FAILURE );
		}
	size = s.st_size;

	sprintf( buffer, "sha1 %d\n", size );
	if( send(server_socket, buffer, strlen(buffer), 0) == INVALID_SOCKET )
		{
		if( errno != EPIPE  &&  errno != ECONNRESET )
			perror_and_exit( "send()" );  /* houve um erro, ou... */
		/* ...o outro lado desligou-se */
		fclose( fp );
		return 0;
		}

	while( size > 0 )
		{
		send_sz = ( size < sizeof(buffer) ? size : sizeof(buffer) );
		send_sz = fread( buffer, 1, send_sz, fp );
		if( send_sz <= 0 )
			break;
		if( send(server_socket, buffer, send_sz, 0) == INVALID_SOCKET )
			{
			if( errno != EPIPE  &&  errno != ECONNRESET )
				perror_and_exit( "send()" );  /* houve um erro, ou... */
			/* ...o outro lado desligou-se */
			fclose( fp );
			return 0;
			}
		size -= send_sz;
		}

	fclose( fp );

	if( size > 0 )
		fprintf( stderr, "** Atencao! Nao consegui enviar os ultimos %d bytes do ficheiro!\n", size );
	return 1;
}


int main( int argc, char *argv[] )
{
	char                *hostname;
	uint16_t            port;
	char                buffer[1024+1], buffer2[1024+1];
	char                *p;
	int                 server_socket;
	struct sockaddr_in  server_sa;
	struct hostent      *hp;
	int                 recv_sz;

#ifdef WIN32
	WSADATA wsaData;
	int wsaErr = WSAStartup( MAKEWORD(2,2), &wsaData );
	if( wsaErr != 0 )
		{
		fprintf( stderr, "Erro %i do WSAStartup() do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms742213(v=vs.85).aspx\npara saber a descricao.\n", (int) wsaErr );
		exit( EXIT_FAILURE );
		}
#else
	signal( SIGPIPE, SIG_IGN );
#endif

	/* Establecer o hostname (FQDN) e a porta IP para o servidor
	*/
	hostname = (argc > 1 ? argv[1] : DEFAULT_HOST );
	port = DEFAULT_PORT;
	p = strchr( hostname, ':' );
	if( p != NULL )
		{
		port = (uint16_t) atol( p+1 );
		*p = '\0';
		}
	if( port < IPPORT_USERRESERVED )
		{
		fprintf( stderr, "Porta incorrecta (%u). Deveria ser >= a %u para evitar conflitos.\n", port, IPPORT_USERRESERVED );
		exit( EXIT_FAILURE );
		}

	/* Determinar o IPv4 do hostname (FQDN)
	*/
	hp = gethostbyname( hostname );
	if( hp == NULL )
		perror_and_exit( "gethostbyname()" );
	if( hp->h_addrtype == AF_INET )
		{
		memset( &server_sa, 0, sizeof(server_sa) );
		server_sa.sin_family = AF_INET;
		memcpy( &server_sa.sin_addr, hp->h_addr, hp->h_length );
		server_sa.sin_port = htons( port );
		}
	else
		{
		fprintf( stderr, "O servidor '%s' tem um endereco de uma familia desconhecida.\n", hostname );
		exit( EXIT_FAILURE );
		}

	/* Obter um socket (uma "tomada") para falar IPv4
	*/
	server_socket = socket( PF_INET, SOCK_STREAM, 0 );
	if( server_socket == INVALID_SOCKET )
		perror_and_exit( "socket()" );

	/* Estabelecer a ligacao
	*/
	if( connect(server_socket, (struct sockaddr*) &server_sa, sizeof(server_sa)) == SOCKET_ERROR )
		perror_and_exit( "connect()" );

	/* Informar o utilizador de onde esta' este servidor ligado
	*/
	if( server_sa.sin_family == AF_INET )
		{
		printf( "** Liguei-me ao servidor '%s' no endereco IPv4 %s:%u...\n",
			hostname,
			inet_ntoa(server_sa.sin_addr), ntohs(server_sa.sin_port) );
		}
	else
		printf( "** Liguei-me ao servidor '%s' numa familia de enderecos que nao IPv4...\n",
			hostname );
	puts( "** Escreve EOF (Ctrl+D ou Ctrl+Z) para me forcar a desligar deste servidor.\n"
	      "** Escreve '*sha1 NOME-DO-FICHEIRO' para enviar o conteudo desse ficheiro para\n"
	      "   o servidor, num comando 'sha1 TAMANHO-DO-FICHEIRO...'.\n"
	      "** Repara como o comando 'sha1 TAMANHO-DO-FICHEIRO...' nao funciona directamente,\n"
	      "   porque o cliente fica 'a espera do servidor apos o comando.\n" );

	for(;;)
		{
		/* Ler mensagem do servidor
		*/
		recv_sz = recv( server_socket, buffer, sizeof(buffer), 0 );
		if( recv_sz == SOCKET_ERROR )
			{
			if( errno != EPIPE  &&  errno != ECONNRESET )
				perror_and_exit( "recv()" );  /* houve um erro, ou... */
			/* ...o outro lado desligou-se */
			recv_sz = 0;
			}
		if( recv_sz == 0 )
			{
			puts( "\n** O servidor desligou a ligacao." );
			break;  /* EOF = o servidor desligou */
			}
		buffer[recv_sz] = '\0';  /* assegurar-me que buffer e' uma string em C */
		fputs( buffer, stdout );

		/* Ler linha do teclado para enviar para o servidor
		*/
		p = fgets( buffer, sizeof(buffer), stdin );
		if( p == NULL )
			{
			if( feof(stdin) )
				break;
			perror( "fgets()" );
			exit( EXIT_FAILURE );
			}

		/* Testa comandos locais especiais (so' um por enquanto);
		   Se nao for especial, envia a linha para o servidor
		*/
		assert( sizeof(buffer) == sizeof(buffer2) );
		if( sscanf(buffer, "*sha1 %[^\r\n]", buffer2) == 1 )
			{
			if( !send_file(server_socket, buffer2) )
				{
				puts( "\n** O servidor desligou a ligacao." );
				break;
				}
			}
		else
			{
			if( send(server_socket, buffer, strlen(buffer), 0) == INVALID_SOCKET )
				{
				if( errno != EPIPE  &&  errno != ECONNRESET )
					perror_and_exit( "send()" );  /* houve um erro, ou... */
				/* ...o outro lado desligou-se */
				puts( "\n** O servidor desligou a ligacao." );
				break;
				}
			}
		}

	/* Fechar a ligacao e terminar
	*/
       	shutdown( server_socket, 2 );
#ifdef WIN32
	WSACleanup();
#endif
	return EXIT_SUCCESS;
}
